/*
 * Test_HW9UnitTest.hpp
 *
 *  Created on: Mar 29, 2015
 *      Author: Joost Huizinga
 */

#ifndef TEST_HW9UNITTEST_HPP_
#define TEST_HW9UNITTEST_HPP_

#include "Ind_CPPN.hpp"

/**
 * Activates the network.
 */
void activate(CPPN& network){
    size_t depth = network.getDepth();
    for(size_t i=0; i<depth; ++i){
        network.step();
    }
}


/**
 * Compares two vectors, prints them, and marks errors.
 */
template<typename Value>
bool compare(std::vector<Value> result, std::vector<Value> expected, bool verbose = false, std::string test = ""){
    bool correct = true;
    for(size_t i = 0; i<result.size(); ++i){
        if(result[i] != expected[i]){
            correct = false;
        }
    }

    if(!correct || verbose){
        std::cout << "Output: ";
        for(size_t i = 0; i<result.size(); ++i){
            std::cout << result[i] << " ";
            if(result[i] != expected[i]){
                std::cout << " (error!) ";
            }
        }
        std::cout << std::endl;

        std::cout << "Expect: ";
        for(size_t i = 0; i<expected.size(); ++i){
            std::cout << expected[i] << " ";
            if(result[i] != expected[i]){
                std::cout << " (error!) ";
            }
        }
        std::cout << std::endl;
    }

    if(test != ""){
        if(correct){
            std::cout << test << ": test success" << std::endl;
        } else {
            std::cout << "WARNING " << test << ": test failed" << std::endl;
        }
    }

    return correct;
}

bool compare(double result, double expected, bool verbose = false, std::string test = ""){
    std::vector<double> result_v;
    std::vector<double> expected_v;
    result_v.push_back(result);
    expected_v.push_back(result);
    return compare(result_v, expected_v, verbose, test);
}

bool compare(size_t result, size_t expected, bool verbose = false, std::string test = ""){
    std::vector<size_t> result_v;
    std::vector<size_t> expected_v;
    result_v.push_back(result);
    expected_v.push_back(result);
    return compare(result_v, expected_v, verbose, test);
}

/**
 * Function useful for asserting that your implementation of CPPN NEAT is working correctly.
 */
void checkCPPN(bool verbose = false){
        CPPN test(2, 1);
        std::vector<double> result;
        std::vector<double> expected;

        //Test a simple linear activation function
        result.clear();
        expected.clear();
        test.getConnections()[0].setWeight(1);
        test.getConnections()[1].setWeight(-1);

        std::cout << "Testing simple network, linear activation function" << std::endl;
        if(verbose){
            test.debugPrint();
        }
        compare(test.getDepth(), 1, verbose, "Testing depth");

        for(double y = -1; y<=1; y+=0.5){
            result.clear();
            expected.clear();
            for(double x = -1; x<=1; x+=0.1){

                test.setValue(0, x);
                test.setValue(1, y);
                size_t depth = test.getDepth();
                for(size_t i=0; i<depth; ++i){
                    test.step();
                }
                result.push_back(test.getOutputValue(0));
//                std::cout << test.getOutputValue(0) << " ";
            }
//            std::cout << std::endl;
//            std::cout << "Expect: ";
            for(double x = -1; x<=1; x+=0.1){
                expected.push_back(std::max(std::min(x - y, 1.0), -1.0));
//                std::cout << std::max(std::min(x - y, 1.0), -1.0) << " ";
            }
            compare(result, expected);
//            std::cout << std::endl;
        }

        if(compare(result, expected, verbose)){
            std::cout << "Test success" << std::endl;
        } else {
            std::cout << "Test failed" << std::endl;
        }

        //Test Gaussian activation function
        result.clear();
        expected.clear();

        test.reset();
        test.getNeurons()[2].setActivationFunction(CPPN_Neuron::gaussian);
        test.setValue(1, 0);

        std::cout << std::endl;
        std::cout << "Testing Gaussian activation function" << std::endl;
        if(verbose){
            test.debugPrint();
        }

//        std::cout << "Output: ";
        for(double x = -1; x<=1; x+=0.2){
            test.setValue(0, x);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            result.push_back(test.getOutputValue(0));
//            std::cout << test.getOutputValue(0) << " ";
        }
//        std::cout << std::endl;
//        std::cout << "Expect: ";
        for(double x = -1; x<=1; x+=0.2){
            expected.push_back(std::exp(float(-x *x)) * 2.0 - 1.0);
//            std::cout << std::exp(float(-x *x)) * 2.0 - 1.0 << " ";
        }
        if(compare(result, expected, verbose)){
            std::cout << "Test success" << std::endl;
        } else {
            std::cout << "Test failed" << std::endl;
        }
//        std::cout << std::endl;


        //Test Sin activation function
        result.clear();
        expected.clear();

        test.reset();
        test.getNeurons()[2].setActivationFunction(CPPN_Neuron::sine);
        test.setValue(1, 0);

        std::cout << std::endl;
        std::cout << "Testing sin activation function" << std::endl;
        if(verbose){
            test.debugPrint();
        }

//        std::cout << "Output: ";
        for(double x = -1; x<=1; x+=0.2){
            test.setValue(0, x);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            result.push_back(test.getOutputValue(0));
//            std::cout << test.getOutputValue(0) << " ";
        }
//        std::cout << std::endl;
//        std::cout << "Expect: ";
        for(double x = -1; x<=1; x+=0.2){
            expected.push_back(std::sin(x));
//            std::cout << std::sin(x) << " ";
        }
//        std::cout << std::endl;
        if(compare(result, expected, verbose)){
            std::cout << "Test success" << std::endl;
        } else {
            std::cout << "Test failed" << std::endl;
        }


        //Test Sigmoid activation function
        result.clear();
        expected.clear();

        test.reset();
        test.getNeurons()[2].setActivationFunction(CPPN_Neuron::sigmoid);
        test.setValue(1, 0);

        std::cout << std::endl;
        std::cout << "Testing sigmoid activation function" << std::endl;
        if(verbose) test.debugPrint();


//        std::cout << "Output: ";
        for(double x = -1; x<=1; x+=0.2){
            test.setValue(0, x);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            result.push_back(test.getOutputValue(0));
//            std::cout << test.getOutputValue(0) << " ";
        }
//        std::cout << std::endl;
//        std::cout << "Expect: ";
        for(double x = -1; x<=1; x+=0.2){
            expected.push_back(tanh(x * 5.0));
//            std::cout << tanh(x * 5.0) << " ";
        }
//        std::cout << std::endl;
        if(compare(result, expected, verbose)){
            std::cout << "Test success" << std::endl;
        } else {
            std::cout << "Test failed" << std::endl;
        }


        //Test activations for larger network
        test.reset();
        test.getNeurons()[2].setActivationFunction(CPPN_Neuron::linear);
        test.addNeuron();
        test.addNeuron();
        test.addNeuron();
        test.getConnections()[0].setWeight(0);
        test.getConnections()[1].setWeight(0);
        test.addConnection(0, 5, 1);
        test.addConnection(5, 4, 1);
        test.addConnection(4, 2, 1);
        test.addConnection(1, 3, 1);
        test.addConnection(3, 2, 1);


        std::cout << std::endl;
        std::cout << "Testing multi-layered network" << std::endl;
        if(verbose) test.debugPrint();
        compare(test.getDepth(), 3, verbose, "Testing depth");

        for(double y = -1; y<=1; y+=0.5){
            result.clear();
            expected.clear();
//            std::cout << "Output: ";
            for(double x = -1; x<=1; x+=0.1){
                test.setValue(0, x);
                test.setValue(1, y);
                size_t depth = test.getDepth();
                if(verbose) std::cout << "Activation values:" << std::endl;
                for(size_t i=0; i<depth; ++i){
                    test.setValue(0, x);
                    test.setValue(1, y);
                    if(verbose){
                        std::cout << "Step " << i << ": ";
                        for(int z=0; z<6; ++z){
                            std::cout  << test.getValue(z) << " ";
                        }
                        std::cout << std::endl;
                    }
                    test.step();
                }
                if(verbose){
                    std::cout << "Step " << depth << ": ";
                    for(int z=0; z<6; ++z){
                        std::cout << test.getValue(z) << " ";
                    }
                    std::cout << std::endl;
                }
                result.push_back(test.getOutputValue(0));


            }
//            std::cout << std::endl;

//            std::cout << "Expect: ";
            for(double x = -1; x<=1; x+=0.1){
                expected.push_back(std::max(std::min(x + y, 1.0), -1.0));
//                std::cout << std::max(std::min(x + y, 1.0), -1.0) << " ";
            }

            compare(result, expected, verbose, "Comparing output");
//            std::cout << std::endl;
        }
}

/**
 * Useful for testing the randomization and mutation properties of the network.
 */
void testRandomization(){
    seed(1);
    CPPN test(2, 1);
    test.reset();
    test.getNeurons()[2].setActivationFunction(CPPN_Neuron::linear);
    test.addNeuron();
    test.addNeuron();
    test.addNeuron();
    test.getConnections()[0].setWeight(0);
    test.getConnections()[1].setWeight(0);
    test.addConnection(0, 5, 1);
    test.addConnection(5, 4, 1);
    test.addConnection(4, 2, 1);
    test.addConnection(1, 3, 1);
    test.addConnection(3, 2, 1);

    //Test randomization
    std::cout << std::endl;
    std::cout << "Randomization test" << std::endl;
    test.randomize();
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    test.randomize();
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    test.randomize();
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    test.randomize();
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;

    //Test mutation
    std::cout << std::endl;
    std::cout << "Mutation test" << std::endl;
    for(size_t i=0; i<20; ++i){
        test.mutate();
    }
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    for(size_t i=0; i<20; ++i){
        test.mutate();
    }
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    for(size_t i=0; i<20; ++i){
        test.mutate();
    }
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
    std::cout << std::endl;
    for(size_t i=0; i<20; ++i){
        test.mutate();
    }
    test.debugPrint();
    std::cout << "Output: ";
    for(double y = -1; y<=1; y+=0.5){
        for(double x = -1; x<=1; x+=0.5){
            test.setValue(0, x);
            test.setValue(1, y);
            size_t depth = test.getDepth();
            for(size_t i=0; i<depth; ++i){
                test.step();
            }
            std::cout << test.getOutputValue(0) << " ";
        }
    }
}



#endif /* TEST_HW9UNITTEST_HPP_ */
